      subroutine abmp16(xb,q2,PDFS,DPDFS,NFLV,IPAR,IORD)
c
c     This is a code for the 3-, 4-, and 5-flavour NNLO nucleon parton 
c     distributions generated from the fit of hep-ph/16xx.xxxxx
c     using the matching conditions of [Eur.Phys.J.C1:301-320,1998], 
c     with account of the PDF uncertainties. 
c
c     The Q**2 range is 0.8d0 < Q**2 < 2d8, the x range is 1d-7 < x < 1d0. 
c
c  Input parameters:
c    XB is the parton momentum fraction 
c    Q2 is the factorization scale
c    NFLV selects the PDFs sets with different number of flavours (3, 4, or 5)
c    IPAR controls output of the PDFs and \alpha_s uncertaintis (see 
c         description of the output parameters below)
c    IORD defines the order of PDFs (2: NNLO, 1: NLO)

c  Output parameters:
c     The array PDFS contains fitted values of the strong coupling constant 
c     and the parton distributions at given x and Q:
c        PDFS(-3) -- mc
c        PDFS(-2) -- mb
c        PDFS(-1) -- mt
c        PDFS(0) -- \alpha_s
c        PDFS(1) -- valence u-quarks 
c        PDFS(2) -- valence d-quarks
c        PDFS(3) -- gluons 
c        PDFS(4) -- sea u-quarks 
c        PDFS(5) -- s-quarks 
c        PDFS(6) -- sea d-quarks 
c        PDFS(7) -- c-quarks
c        PDFS(8) -- b-quarks
c     Output array DPDFS(0:8,NVAR) contains derivatives of \alpha_s and
c     the PDFs on the parameters corresponding to the independent 
c     sources of the uncertainties, NVAR is the number of these sources, 
c     equal to 28 in the current version. The input parameter IPAR is used to 
c     optimize performance fo the code. If IPAR=0, no 
c     uncertainties are returned in DPDFS; if 0<IPAR<=NVAR, only 
c     the uncertainty due to the IPAR-th source is returned; if IPAR<0
c     all uncertainties for the sources from 1 to NVAR are returned. 
c     Using derivatives returned in DPDFS one can take into account the 
c     correlations between different PDFs and between PDFs and \alpha_s. 
c     All derivatives are transformed to the orthonormal basis of eigenvectors 
c     of the parameters error matrix therefore variation of the PDFs by 
c     the values of DPDFS is performed independently. For example, 
c     after the call with IPAR=-1 the dispersion of the i-th PDF can 
c     be stored in DELPDF using the code 
c
c-----------------
c          DELPDF=0.
c          do k=1,nvar
c            DELPDF=DELPDF+dpdfs(i,k)**2
c          end do
c-----------------
c     and its random value can be stored in RPDF using the code 
c-----------------
c          RPDF=pdfs(i)          
c          do k=1,nvar
c            s=0.
c            do l=1,96
c              s=s+(2*rndm(xxx)-1)/sqrt(32.)
c            end do
c            RPDF=RPDF+s*dpdfs(i,k)
c          end do

      implicit none

      integer nxb,nq,np,nvar
      parameter(nxb=99,nq=20,np=8,nvar=29)

      integer k,i,n,m,kx,nxbb,nxb1,nxb2
      integer NPDF,NFLV

      real*8 f(nxb,nq+1,-3:np),xx(nxb)
      real*8 fsp(nxb),bs(nxb),cs(nxb),ds(nxb)
      real*8 bsp(nxb,nq+1,-3:np),csp(nxb,nq+1,-3:np),dsp(nxb,nq+1,-3:np)
      real*8 bspd(nvar,nxb,nq+1,-3:np),cspd(nvar,nxb,nq+1,-3:np)
     ,      ,dspd(nvar,nxb,nq+1,-3:np)
      real*8 pdfs(-3:np),dpdfs(-3:np,nvar)
      real*8 df(nvar,-3:np,nxb,nq+1)
      real*8 x,qsq,dels,delx,x1,delx1,xlog1,xd,b,aa,ss,f0,fp,fm
      real*8 xb,q2,df0,dfp,dfm

      character pdford*1,suffix*4
      dimension pdford(3),suffix(2)

      real*8 xmin,xmax,xmax1,qsqmin,qsqmax
      integer nflvs,npar1,npar2,ipar,iord
      integer lnblnk

c I/O channel to read the data
      integer nport
      character locdir*128
      data nport/1/

      data pdford/'3','4','5'/
      data suffix/'_nlo','    '/

      data xmin,xmax,qsqmin,qsqmax/1d-7,1d0,0.8d0,2d8/
      data NFLVS /-1/

      save nflvs,f,df,dels,delx,x1,delx1,xlog1,nxbb,xx

c put in your local address of the PDF files in LOCDIR 
      data locdir /' '/
c or in the system variable GRIDS
!      CALL GETENV( 'GRIDS', locdir ) 
      locdir=locdir(:LNBLNK(locdir))//'pdfsets/'

      npdf=nflv+3

      if (ipar.gt.nvar) print *,'Wrong call of the PDF uncertainties' 

      if (ipar.eq.0) then 
        npar1=0
        npar2=0
      end if
      if (ipar.ge.0) then 
        npar1=ipar
        npar2=ipar
      end if
      if (ipar.lt.0) then 
        npar1=1
        npar2=nvar
      end if

      if (nflvs.eq.nflv) goto 10

      nflvs=nflv

      dels=(dlog(dlog(qsqmax/0.04d0))-
     +      dlog(dlog(qsqmin/0.04d0)))/dble(nq-1)

      nxb1=30
      nxb2=nxb-nxb1
      x1=0.3d0
      xmax1=0.99d0
      delx=(log(x1)-log(xmin))/(nxb1-1)
      DELX1=(log(1-xmax1)-log(1-x1))/(nxb2-1)

      do kx=1,nxb1
        xx(kx)=exp(log(xmin)+delx*(kx-1))
      end do

      do kx=nxb-1,nxb1,-1
        xx(kx)=1-exp(log(1-xmax1)+delx1*(kx+1-nxb))
      end do
      xx(nxb)=1d0

*...Read input tables
      print *,'***** Reading PDFs from tables *****'
      open(unit=nport,status='old'
     ,    ,file=locdir(:LNBLNK(locdir))
     /     //'/abmp16.dpdfs_'//pdford(nflv-2)//suffix(iord))

      do n=1,nxb-1
        do m=1,nq
          do i=-3,npdf 
            read (nport,*) (df(k,i,n,m),k=1,nvar)
          end do
        end do
      end do
      close(unit=nport)

      do k=1,nvar
        do i=-3,npdf
          do m=1,nq
            if (i.ne.0) then 
              df(k,i,nxb,m)=0d0
            else 
              df(k,i,nxb,m)=df(k,i,nxb-1,m)
            end if
            do n=1,nxb
              fsp(n)=df(k,i,n,m)
            end do
            call spline (nxb,xx,fsp,bs,cs,ds)
            do n=1,nxb
              bspd(k,n,m,i)=bs(n)
              cspd(k,n,m,i)=cs(n)
              dspd(k,n,m,i)=ds(n)
            end do
          end do
        end do
      end do

      open(unit=nport,status='old',err=199
     ,    ,file=locdir(:LNBLNK(locdir))
     /     //'/abmp16.pdfs_'//pdford(nflv-2)//suffix(iord))
      do n=1,nxb-1
        do m=1,nq
          read(nport,*) (f(n,m,i),i=-3,npdf)
        end do
      end do
      do i=-3,npdf
        do m=1,nq
          if (i.ne.0) then 
            f(nxb,m,i)=0d0
          else 
            f(nxb,m,i)=f(nxb-1,m,i)
          end if
          do n=1,nxb-1
            f(n,m,i)=f(n,m,i)
          end do
          do n=1,nxb
            fsp(n)=f(n,m,i)
          end do
          call spline (nxb,xx,fsp,bs,cs,ds)
          do n=1,nxb
            bsp(n,m,i)=bs(n)
            csp(n,m,i)=cs(n)
            dsp(n,m,i)=ds(n)
          end do
        end do
      end do
      close(unit=nport)

  10  continue

      if((q2.lt.qsqmin).or.(q2.gt.qsqmax)) then
         print 99,q2,qsqmin,qsqmax
         return
      end if
      if((xb.lt.xmin).or.(xb.gt.xmax)) then
         print 98,xb,xmin,xmax
         return
      end if
  99  format('  AGRIDS WARNING:  Q^2 VALUE IS OUT OF RANGE   ',3g12.3)
  98  format('  AGRIDS WARNING:   X  VALUE IS OUT OF RANGE   ',3g12.3)

      x=max(xb,xmin)
      x=min(xb,xmax)
      qsq=max(q2,qsqmin)
      qsq=min(q2,qsqmax)

      do n=1,nxb-1
        if (x.lt.xx(n+1)) goto 300
        end do

 300    aa=x-xx(n)

      ss=dlog(dlog(qsq/0.04d0))-dlog(dlog(qsqmin/0.04d0))
      m=int(ss/dels)+1
      b=ss/dels-dble(m)+1.d0

      do i=-3,npdf
        f0=f(n,m,i) + aa*bsp(n,m,i) + aa**2*csp(n,m,i) 
     +              + aa**3*dsp(n,m,i)
        fp=f(n,m+1,i) + aa*bsp(n,m+1,i) + aa**2*csp(n,m+1,i)
     +                + aa**3*dsp(n,m+1,i)
        if (m.ge.2) then 
          fm=f(n,m-1,i) + aa*bsp(n,m-1,i) + aa**2*csp(n,m-1,i)
     +                   +aa**3*dsp(n,m-1,i)
          pdfs(i)=fm*b*(b-1d0)/2d0 + f0*(1d0-b**2) + fp*b*(b+1d0)/2d0
        else 
          pdfs(i)= f0*(1d0-b) + fp*b
        end if
        if (npar1.gt.0) then 
          do k=npar1,npar2
            df0=df(k,i,n,m) + aa*bspd(k,n,m,i) + aa**2*cspd(k,n,m,i) 
     +                      + aa**3*dspd(k,n,m,i)
            dfp=df(k,i,n,m+1)+aa*bspd(k,n,m+1,i)+aa**2*cspd(k,n,m+1,i)
     +                        + aa**3*dspd(k,n,m+1,i)
            if (m.ge.2) then 
              dfm=df(k,i,n,m-1)+aa*bspd(k,n,m-1,i)+aa**2*cspd(k,n,m-1,i)
     +                          + aa**3*dspd(k,n,m-1,i)
              dpdfs(i,k)=dfm*b*(b-1d0)/2d0 
     +                  + df0*(1d0-b**2) +dfp*b*(b+1d0)/2d0
            else 
              dpdfs(i,k) = df0*(1d0-b) + dfp*b
            end if
          end do
        end if
      end do

      return

 199  print *,'The PDFs set is inavailable'

      return
      end
* ---------------------------------------------------------------------
      SUBROUTINE SPLINE(N,X,Y,B,C,D)
* ---------------------------------------------------------------------
* CALCULATE THE COEFFICIENTS B,C,D IN A CUBIC SPLINE INTERPOLATION.
* INTERPOLATION SUBROUTINES ARE TAKEN FROM
* G.E. FORSYTHE, M.A. MALCOLM AND C.B. MOLER,
* COMPUTER METHODS FOR MATHEMATICAL COMPUTATIONS (PRENTICE-HALL, 1977).
*
      IMPLICIT REAL*8(A-H,O-Z)

      DIMENSION X(N), Y(N), B(N), C(N), D(N)
*
      NM1=N-1
      IF(N.LT.2) RETURN
      IF(N.LT.3) GO TO 250
      D(1)=X(2)-X(1)
      C(2)=(Y(2)-Y(1))/D(1)
      DO 210 K=2,NM1
         D(K)=X(K+1)-X(K)
         B(K)=2.0D0*(D(K-1)+D(K))
         C(K+1)=(Y(K+1)-Y(K))/D(K)
         C(K)=C(K+1)-C(K)
  210 CONTINUE
      B(1)=-D(1)
      B(N)=-D(N-1)
      C(1)=0.0D0
      C(N)=0.0D0
      IF(N.EQ.3) GO TO 215
      C(1)=C(3)/(X(4)-X(2))-C(2)/(X(3)-X(1))
      C(N)=C(N-1)/(X(N)-X(N-2))-C(N-2)/(X(N-1)-X(N-3))
      C(1)=C(1)*D(1)**2.0D0/(X(4)-X(1))
      C(N)=-C(N)*D(N-1)**2.0D0/(X(N)-X(N-3))
 215  CONTINUE
      DO 220 K=2,N
         T=D(K-1)/B(K-1)
         B(K)=B(K)-T*D(K-1)
         C(K)=C(K)-T*C(K-1)
 220  CONTINUE
      C(N)=C(N)/B(N)
      DO 230 IB=1,NM1
         K=N-IB
         C(K)=(C(K)-D(K)*C(K+1))/B(K)
 230  CONTINUE
      B(N)=(Y(N)-Y(NM1))/D(NM1)
     1     +D(NM1)*(C(NM1)+2.0D0*C(N))
      DO 240 K=1,NM1
         B(K)=(Y(K+1)-Y(K))/D(K)
     1        -D(K)*(C(K+1)+2.0D0*C(K))
         D(K)=(C(K+1)-C(K))/D(K)
         C(K)=3.0D0*C(K)
 240  CONTINUE
      C(N)=3.0D0*C(N)
      D(N)=D(N-1)
      RETURN
 250  CONTINUE
      B(1)=(Y(2)-Y(1))/(X(2)-X(1))
      C(1)=0.0D0
      D(1)=0.0D0
      B(2)=B(1)
      C(2)=0.0D0
      D(2)=0.0D0
      RETURN
      END
